using dg1d
using CairoMakie
using LinearAlgebra
CairoMakie.activate!()


function analyze_convergence()

  # inputs

  norm_order = 1
  norm_fn = (err, mesh) -> norm(err, mesh, norm_order)

  type_reference_solution = :analytic
  # sample_resolutions = Dict("Mesh.N" => 1, "Mesh.K" => [16,32,64,128,256,512])
  # sample_resolutions = Dict("Mesh.N" => 2, "Mesh.K" => [16,32,64,128,256,512])
  # sample_resolutions = Dict("Mesh.N" => 3, "Mesh.K" => [16,32,64,128,256,512])
  sample_resolutions = Dict("Mesh.N" => 4, "Mesh.K" => [16,32,64,128,256,512])
  
  thisdir = dirname(@__FILE__)
  table = gather_parameters(thisdir)
  subtable, diffs = match_properties(table, sample_resolutions, ignore=["filename"])

  @info "Samples for convergence rate estimation"
  subtable = sort(subtable, ["Mesh.K", "Mesh.N"])
  display(subtable[:, ["Mesh.N", "Mesh.K", "filename"]])

  @info "Sample differences"
  display(diffs)
  if !(issubset(keys(diffs), [ "Mesh.N", "Mesh.K", "filename", "Mesh.CFL" ]))
    error("Sample data sets contain differences in parameters which spoil the analysis.")
  end

  list_prms      = [ to_dict(FlatDict, df) for df in eachrow(subtable) ]
  list_mesh      = [ make_Mesh(trim_headers(prms, "Mesh")) for prms in list_prms ]
  list_filenames = [ df."filename" for df in eachrow(subtable) ]
	list_indices   = [ get_parameter_file_index(filename) for filename in list_filenames ]

	# setup plot directory
	plotdir = joinpath(thisdir, "plots")
	isfile(plotdir) && error("There is a file where we want to place plots: '$plotdir'")
	!isdir(plotdir) && mkdir(plotdir)
  plotname_postfix = join(list_indices, "_")

  id_prms = trim_headers(first(list_prms), "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_prms)
  # Using Initialdata and Mesh parameters is not ideal: Mesh.K could be different between
  # runs, but we know that make_function_initialdata_equation should not care about Mesh.K
  # Perhaps we need to redefine the meaning of the Mesh section: E.g Mesh refers to the dimensions
  # of the physical domain, whereas grid referes to the layout of the computationl domain?
  ana_parameters = first(list_prms)
  id_parameters = trim_headers(ana_parameters, "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_parameters)
  v = ana_parameters["Equation.v"]
  analytic_u(x, t) = id_fn(x - v * t)

  # Strategy:
  # 0) compute data on aligned time slices (must be before this script is invoked),
  # 1) load data,
  # 2a) compute the analytical solution on the sample's grid,
  # 2b) or interpolate the reference solution (using its own representation) onto the sample's grid;
  # Using linear interpolation (even for spatially very high resolved grids) introduces unnecessary
  # errors, see. https://git.tpi.uni-jena.de/dg/dg1d.jl/-/issues/30
  # 3) compute errors and norms thereof,
  # 4) plot.
  
  list_output = [ Output(joinpath(thisdir, first(splitext(filename)), "output.h5"))
                  for filename in list_filenames ]
  aligned_times = first(list_prms)["Output.aligned_times"]
  list_cvgrates = zeros(Float64, length(list_output)-1, 0)
  for intrp_t in aligned_times
    list_us = [ read_variable(output, "dynamic_variables.u", intrp_t) 
                for output in list_output ]

    list_interpolated_ref_us = if type_reference_solution isa Symbol
      # analytical reference solution
      [ analytic_u.(view(mesh.x,:), intrp_t) for mesh in list_mesh ]
    else
      # numerical reference solution
      @unpack N, K = type_reference_solution # resolution
      subtable_ref, _ = match_properties(table, Dict("Mesh.K"=>K, "Mesh.N"=>N))
      nrow(subtable_ref) == 1 || error("No unique reference solution found!")
      # @info "Numerical reference solution"
      # display(subtable_ref[:,["Mesh.N", "Mesh.K", "filename"]])

      parameters_ref = first(subtable_ref)
      ref_mesh = make_Mesh(trim_headers(to_dict(FlatDict, parameters_ref), "Mesh"))
      make_reference_interpolator(smpl_mesh) = MeshInterpolator(smpl_mesh, ref_mesh)

      ref_output = Output(joinpath(thisdir, first(splitext(parameters_ref."filename")), "output.h5"))

      ref_u                    = read_variable(ref_output, "dynamic_variables.u", intrp_t)
      list_interpolators       = [ make_reference_interpolator(mesh) for mesh in list_mesh ]
      list_interpolated_ref_us = [ view(interpolator(ref_u), :) for interpolator in list_interpolators ]
      list_interpolated_ref_us
    end
    list_errors              = [ u .- intrp_ref_u
                                 for (u, intrp_ref_u) in zip(list_us, list_interpolated_ref_us) ]
    list_error_norms         = [ norm_fn(error, mesh) 
                                 for (error,mesh) in zip(list_errors, list_mesh) ]
    cvgrates = map(1:length(list_us)-1) do idx
      err1, err2 = list_error_norms[idx], list_error_norms[idx+1]
      mesh1, mesh2 = list_mesh[idx], list_mesh[idx+1]
      NpK1, NpK2 = length(mesh1), length(mesh2)
      estimate_convergence_rate(NpK1, NpK2, err1, err2)
    end

    list_cvgrates = hcat(list_cvgrates, cvgrates)

    fig = Figure(resolution=(1600,1200))
    ax1 = Axis(fig[1,1], ylabel=L"\log(|u - u_{ref}|)", xlabel=L"x",
               title="Differences wrt reference solution @ t = $intrp_t")
    # set artificial ylimits to prevent autolimits! choking on invalid y log-range if errors are zero
    for (i, (error, mesh)) in enumerate(zip(list_errors, list_mesh))
      L = layout(mesh)
      x_w_nans = make_nan_separated_storage(L)
      x_w_nans[1:end-1,:] .= mesh.x
      error_w_nans = make_nan_separated_storage(L)
      error_w_nans[1:end-1,:] .= abs.(reshape(error, L))
      log_error_w_nans = log10.(error_w_nans)
      Npts, K = L
      lines!(ax1, x_w_nans[:], log_error_w_nans[:], label="($Npts, $K)")
    end

    axislegend("Resoluion: (Npts, K)")

    plotdir = joinpath(thisdir, "plots")
    filename = "difference_t_$(intrp_t)_$(plotname_postfix).png"
    @info "Saving error plot to '$filename'"
    save(joinpath(plotdir, filename), fig)

  end

  @info "Convergence rates"
  display(list_cvgrates)

  # plot convergence rates
  fig       = Figure(resolution=(1600,1200))
  subfig1   = fig[1,1]
  ax1       = Axis(subfig1, ylabel= "Convergence rate in |u|_{$(norm_order)} norm", xlabel="t")
  ax1.title = "Evolution of convergence rate"

  for (idx, cvgrate) in enumerate(eachrow(list_cvgrates))
    mesh1, mesh2 = list_mesh[idx], list_mesh[idx+1]
    Npts1, K1 = layout(mesh1)
    Npts2, K2 = layout(mesh2)
    lines!(ax1, aligned_times, cvgrate, 
           label="(Npts_1, K_1) - (Npts_2, K_2) = ($Npts1, $K1) - ($Npts2, $K2)")
  end

  # style
  min_Npts = minimum([ m.element.Npts for m in list_mesh ])
  max_Npts = maximum([ m.element.Npts for m in list_mesh ])
  min_Npts, max_Npts = floor(min_Npts), ceil(max_Npts)
  ylims!(min_Npts-1.5,max_Npts+0.5)

	axislegend(ax1)

  filename = "convergence_rate_files_$(plotname_postfix).png"
  @info "Saving convergence rate plot to '$filename'"
  filename = joinpath(plotdir, filename)
	@time save(filename, fig)

  return
end


#######################################################################
#                                Main                                 #
#######################################################################


analyze_convergence()
