export SphericalAccretion


"""
Solution taken from Bugner [2017], section 4.2.

Kerr-Schild coordinates allow use to go beyond horizon towards the singularity.
However, it becomes more difficult for the Newton Rapshon solver to determine a solution
once beyond the horizon. To this end, try increasing `N_rs` whenever you lower `rmin` and 
the solver aborts inside the horizon.
Alternatively, one could also add an adaptive stepper.

Note: We implicitly assume θ = π/2, because then sin(θ) = 1, although,
the sines should cancel out everywhere.
"""
module SphericalAccretion


using dg1d
using Plots
using Printf


function solve()

  # parameters
  M = 1.0
  N_rs = 100000
  rmin, rmax = 0.1*M, 200*M
  rc, ρc, Γ = 200*M, 7e-4, 4/3

  urc = sqrt(M / (2 * rc))
  # cannot find solutions within these regions, cf. Michel 1972
  @show urc
  @assert rc >= 6*M
  @assert urc <= 1 / 3

  Vc2 = urc^2 / (1 - 3 * urc^2)
  K = Vc2 * (Γ - 1) / ( (Γ - 1 - Vc2) * Γ * ρc^(Γ - 1) )

  # here we set θ = π/2 so that sin(θ) = 1
  C1 = rc^2 * ρc * urc
  hc = 1 + K * Γ * ρc^(Γ - 1) / (Γ - 1)
  u0c = sqrt(1 - 2 * M / rc + urc^2)
  C3 = hc^2 * u0c^2

  function F_dF(ur, r)
    ρ = C1 / (r^2 * ur)
    h = 1 + K * Γ * ρ^(Γ - 1) / (Γ - 1)
    u0 = sqrt(1 - 2 * M / r + ur^2)
    F = h^2 * u0^2 - C3
    dρdur = - C1 / (r^2 * ur^2)
    dhdρ = K * Γ * ρ^(Γ - 2)
    du0dur = ur / u0
    dF = 2 * (u0^2 * h * dhdρ * dρdur + h^2 * u0 * du0dur)
    return F, dF
  end

  function F(ur, r)
    ρ = C1 / (r^2 * ur)
    h = 1 + K * Γ * ρ^(Γ - 1) / (Γ - 1)
    u0 = sqrt(1 - 2 * M / r + ur^2)
    u0 = sqrt(1 - 2 * M / r + ur^2)
    return h^2 * u0^2 - C3
  end

  nlsolve_parmeters = dg1d.NlSolveParameters(maxits=20)

  # radial domain over which we solve
  rs = collect(LinRange(rmin, rmax, N_rs))
  urs = zeros(Float64, N_rs)
  # solve far field first and walk inwards, this way we can use urc as a guess (if rc > rmax)
  for (idx, r) in Iterators.reverse(enumerate(rs))
    # interestingly, the following guesses converge to the wrong solution:
    # - use urc as a guess for the outermost radius
    # - use the previous result for the remaining radi
    # The following converges to the correct solution:
    if r > 2*M
      ur_guess = sqrt(1 - 2 * M / r) # asymptotic value for ur
    else
      ur_guess = urs[idx+1] # use previous iterate where asymptotic value is not defined
    end
    # F_at_r(ur) = F(ur, r)
    # ur, err, converged = dg1d.nlsolve(F_at_r, ur_guess, prms=nlsolve_parmeters)
    F_dF_at_r(ur) = F_dF(ur, r)
    ur, err, converged = dg1d.nlsolve_exact(F_dF_at_r, ur_guess, prms=nlsolve_parmeters)
    if !converged
      println()
      error("Could not determine solution for 'ur' at radius r = $r. " *
            "Remaining error: $err")
    end
    # @show (ur, r)
    if ur < 0.0
      @warn "Aborting due to ur becoming negative at r = $r." 
      @info "$(N_rs - idx) points skipped ..."
      break
    end
    urs[idx] = ur
  end

  ρs = @. C1 / (rs^2 * urs)
  ps = @. K * ρs^Γ
  ϵs = @. ps / ((Γ - 1) * ρs)
  βr = @. 2 * M / (rs + 2 * M)
  α = @. sqrt( rs / (rs + 2 * M) )

  gtt_uu = @. - (2 * M + rs) / rs
  gtr_uu = @. 2 * M / rs
  grr_uu = @. (rs - 2 * M) / rs
  u0s_d = @. sqrt(1 - 2 * M / rs + urs^2)
  u0s_u = @. gtt_uu * u0s_d + gtr_uu / grr_uu * (urs - gtr_uu * u0s_d)
  vrs = @. urs / (α * u0s_u) + βr / α

  Plots.gr()

  vars = [ ρs, ps, ϵs, vrs ]
  ylabels = [ "ρ", "p", "ϵ", "v_r" ]

  plts = [ Plots.plot(xlabel="r", ylabel=ylabels[i], xlimits=extrema(rs))
           for i = 1:length(vars) ]
  for (plt, var) in zip(plts, vars)
    isnothing(var) && continue
    Plots.plot!(plt, rs, var, label=false)
  end

  layout = (Int64(length(vars)/2),2)
  plt = Plots.plot(plts..., layout=layout, plot_title="""
        Spherical accretion, $(@sprintf "M = %.5f, ρc = %.5f, rc = %.5f, Γ = %.5f" M ρc rc Γ)
        """, 
        size=(1400,1000))

  Plots.gui(plt)

  return
end


end # module
