"""
    HybridEoS{T_OneParameterEos} <: AbstractEquationOfState

A thermally hybridized equation of state.
See Shibata et al. 2005, Thierfelder et al. 2011.

  p(ρ,ϵ)      = p^cold(ρ) + (Γ - 1) ρ ϵ^hot
  ϵ^hot(ρ, ϵ) = ϵ - ϵ^cold(ρ)

- `Γ` thermal constant
"""
struct HybridEoS{T_ColdEoS} <: HotEquationOfState
  cold_eos::T_ColdEoS
  gamma::Float64
end


#######################################################################
#                       Thermodynamic variables                       #
#######################################################################


# p = p(ρ,ϵ)
function (eos::HybridEoS)(::Type{Pressure},
                          ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  p_cold = eos.cold_eos(Pressure, Density, rho, InternalEnergy, NaN)
  eps_cold = eos.cold_eos(InternalEnergy, Density, rho, Pressure, NaN)
  eps_hot = eps - eps_cold
  return p_cold + (eos.gamma - 1) * rho * eps_hot
end


# ρ = ρ(ϵ,p)
function (eos::HybridEoS)(::Type{Density},
                         ::Type{InternalEnergy}, eps, ::Type{Pressure}, p)

  TODO(eos)
end


# ϵ = ϵ(ρ,p)
function (eos::HybridEoS)(::Type{InternalEnergy},
                         ::Type{Density}, rho, ::Type{Pressure}, p)

  TODO(eos)
end


# s = s(ρ,ϵ)
function (eos::HybridEoS)(::Type{Entropy},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  TODO(eos)
end


# χ = ∂p/∂ρ(ρ,ϵ)
function (eos::HybridEoS)(::Type{Derivative{Pressure, Density}},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  p_cold = eos.cold_eos(Pressure, Density, rho)
  eps_cold = eos.cold_eos(InternalEnergy, Density, rho)
  eps_hot = eps - eps_cold
  chi_cold = eos.cold_eos(Derivative{Pressure,Density}, Density, rho)
  return chi_cold - (eos.gamma - 1) * (eps_hot - p_cold / rho)
end


# κ = ∂p/∂ϵ(̢ρ,ϵ)
function (eos::HybridEoS)(::Type{Derivative{Pressure, InternalEnergy}},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  return (eos.gamma - 1.0) * rho
end
