#######################################################################
#                            Flux methods                             #
#######################################################################



@with_signature function flux(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy(args, equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho = q
  flx_q   = q^2 / rho + p
  flx_E   = q / rho * (E + p)
  @returns flx_rho, flx_q, flx_E
end



@with_signature function speed(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy(args, equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function internal_energy(::EulerEquation)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, eps
  @unpack eos = equation
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function entropy(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, q, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation)
  @accepts rho, q, S
  @unpack eos  = equation
  flx_S = q / rho * S
  @returns flx_S
end


#######################################################################
#                         LDG implementation                          #
#######################################################################


@with_signature function ldg_flux(equation::EulerEquation)
  @accepts State(rho, q, E)
  flx_ldg_rho, flx_ldg_q, flx_ldg_E = rho, q, E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function ldg_speed(equation::EulerEquation)
  @accepts State(rho, q, E)
  v_ldg_rho, v_ldg_q, v_ldg_E = 1, 1, 1
  @returns v_ldg_rho, v_ldg_q, v_ldg_E
end


@with_signature function av_flux(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), smoothed_mu
  flx_ldg_rho, flx_ldg_q, flx_ldg_E = 
    smoothed_mu * ldg_rho, smoothed_mu * ldg_q, smoothed_mu * ldg_E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function av_speed(equation::EulerEquation)
  @accepts State(flx_ldg_rho, flx_ldg_q, flx_ldg_E), smoothed_mu
  v_av_rho, v_av_q, v_av_E = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_av_rho, v_av_q, v_av_E
end
