#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                   = env
  @unpack equation, rsolver             = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack lhs_numflx_rho, rhs_numflx_rho,
          lhs_numflx_q  , rhs_numflx_q  ,
          lhs_numflx_E  , rhs_numflx_E  = get_bdry_variables(cache)

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::AbstractReconstruction, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                    = env
  @unpack equation, rsolver              = P
  @unpack rho, q, E                      = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps  = get_static_variables(cache)
  @unpack flag                           = get_cell_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E          = get_rhs_variables(cache)
  @unpack lhs_numflx_rho, rhs_numflx_rho,
          lhs_numflx_q  , rhs_numflx_q  ,
          lhs_numflx_E  , rhs_numflx_E   = get_bdry_variables(cache)

  HRSC.reconstruct!(rho,  flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(q,    flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(E,    flag, hrsc, isperiodic=isperiodic(bdryconds))

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver, ldg_rsolver, av_rsolver = P

  @unpack rho, q, E                         = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
                                        ldg_rho, ldg_q, ldg_E,
          flx_ldg_rho, flx_ldg_q, flx_ldg_E = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E             = get_rhs_variables(cache)
  @unpack lhs_numflx_rho,     rhs_numflx_rho, 
          lhs_numflx_q,       rhs_numflx_q,
          lhs_numflx_E,       rhs_numflx_E,
          lhs_numflx_ldg_rho, rhs_numflx_ldg_rho,
          lhs_numflx_ldg_q,   rhs_numflx_ldg_q,
          lhs_numflx_ldg_E,   rhs_numflx_ldg_E  = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(central_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_rho, flx_ldg_rho,  lhs_numflx_ldg_rho, rhs_numflx_ldg_rho, mesh)
  compute_rhs_weak_form!(ldg_q,   flx_ldg_q,    lhs_numflx_ldg_q,   rhs_numflx_ldg_q,   mesh)
  compute_rhs_weak_form!(ldg_E,   flx_ldg_E,    lhs_numflx_ldg_E,   rhs_numflx_ldg_E,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  @. flx_rho        += flx_ldg_rho
  @. flx_q          += flx_ldg_q
  @. flx_E          += flx_ldg_E
  @. lhs_numflx_rho += lhs_numflx_ldg_rho
  @. lhs_numflx_q   += lhs_numflx_ldg_q
  @. lhs_numflx_E   += lhs_numflx_ldg_E
  @. rhs_numflx_rho += rhs_numflx_ldg_rho
  @. rhs_numflx_q   += rhs_numflx_ldg_q
  @. rhs_numflx_E   += rhs_numflx_ldg_E

  compute_rhs_weak_form!(rhs_rho, flx_rho,  lhs_numflx_rho, rhs_numflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    lhs_numflx_q,   rhs_numflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    lhs_numflx_E,   rhs_numflx_E,   mesh)

  return
end
