@with_signature function cons2prim(equation::Equation{<:HotEquationOfState})
  @accepts D, Sr, tau, p, r, B
  S2   = (Sr/B)^2
  @unpack eos, cons2prim_newtonsolver, cons2prim_bisection = equation
  p, converged = find_pressure_insane(D, S2, tau, p, r, eos,
                                      cons2prim_newtonsolver, cons2prim_bisection)
  if !converged && eltype(equation) == HybridEoS
    rho, success = find_density_insane(D, S2, p, r, equation, eos,
                                       cons2prim_newtonsolver, cons2prim_bisection)
    if !success
      error()
    end
    p    = eos(Pressure, Density, rho)
    vr   = Formulae.vr(B, D, Sr, tau, p)
    eps  = Formulae.ϵ(B, D, Sr, tau, p)
  else
    rho  = Formulae.ρ(B, D, Sr, tau, p)
    vr   = Formulae.vr(B, D, Sr, tau, p)
    eps  = Formulae.ϵ(B, D, Sr, tau, p)
  end
  @returns rho, vr, eps, p
end


@with_signature function cons2prim(equation::Equation{<:ColdEquationOfState})
  @accepts D, Sr, tau, p, r, B
  S2   = (Sr/B)^2
  rho, success  = find_density_insane(D, S2, p, equation)
  @unpack eos = equation
  p    = eos(Pressure, Density, rho)
  vr   = Formulae.vr(B, D, Sr, tau, p)
  eps  = Formulae.ϵ(B, D, Sr, tau, p)
  @returns rho, vr, eps, p
end


#######################################################################
#           Insane version - bail out if anything is wrong            #
#######################################################################


function find_pressure_insane(D, S2, tau, p, r, eos::HotEquationOfState,
  cons2prim_newtonsolver, cons2prim_bisection)

  # upper limit on particle velocity
  # omv2 = 1e-3 # = 1 - v^2

  # pmin1 = - (tau + D) + sqrt(S2 + omv2) # √((τ + D + p)^2 - S^2) >= √(1 - vmax^2)
  pmin1 = - (tau + D) + sqrt(S2) + 1e-10 # √((τ + D + p)^2 - S^2) >= √(1 - vmax^2)
  pmin2 = 1e-10 # physical constraint
  pmin = max(pmin1, pmin2)

  floor_pressure(pnew) = max(pnew, pmin)
  # p0 = 0.5 * (eos.gamma - 1) * (tau + D)
  p0 = p

  f = pp -> pressure_equation_insane(D, S2, tau, pp, r, eos)
  p, err, converged, tol = find_root(f, p0, cons2prim_newtonsolver, sanitize=floor_pressure)
  if !converged
    # reset to guess
    p = p0
  end
  p = max(pmin, p0)

  return p, converged
end


function pressure_equation_insane(D, S2, tau, p, r, eos)

  a = tau + p + D
  a2mS2 = a^2 - S2
  if a2mS2 < 0
    return NaN
  end
  W = a / sqrt(a2mS2)
  rho = D / W
  # if rho < 0 #1e-15
  #   # return NaN
  #   # rho = 1e-10
  # end
  eps_rho = (a - W^2 * p) / W^2 - rho
  if eps_rho < 0
    # in low temperature limit we should switch to one parameter EoS
    return NaN
  end
  eps = eps_rho / rho
  eos_p = eos(Pressure, Density, rho, InternalEnergy, eps) + 1e-10
  f = eos_p - p
  return f
end


function find_density_insane(D, S2, rho, eos,
    cons2prim_newtonsolver, cons2prim_bisection)

  rhomin = 1e-10
  rho_init = rho

  floor_density(rhonew) = max(rhonew, rhomin)
  g = rho -> density_equation(D, S2, rho, eos)
  rho, err, converged, tol = find_root(g, rho_init, cons2prim_newtonsolver,
                                       sanitize=floor_density)
  if !converged
    # try bisection
    # Figure out brackets for a bisection method.
    # if bisection failed reset
    rho = rho_init
  end

  success = converged

  return rho, success
end


function density_equation(D, S2, rho, eos)
  D2 = D^2
  h = eos(Enthalpy, Density, rho)
  W = sqrt(1 + S2 / (D2 * h^2))
  chi = eos(Derivative{Pressure,Density}, Density, rho)
  g = W * rho - D
  # assuming zero temperature limit: deps/drho = P / rho^2
  dg = W - S2 * chi / (W * D2 * h^3)
  return g, dg
end
