function initialdata!(env, P::Project, prms)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms)
  # initialdata_entropy_variables!(env, P, prms)
  initialdata_hrsc!(env, P, P.hrsc, prms)
  initialdata_tci!(env, P, P.tci, prms)
  return
end


#######################################################################
#                   Initial data equation of state                    #
#######################################################################


function initialdata_equation_of_state!(env, P::Project, eos::EquationOfState.IdealGas, prms)

  @unpack id = prms
  @unpack cache, mesh = env
  x = vec(mesh.x)

  ρ0, v0, p0 = initialdata_equation_of_state(Val(Symbol(id)), eos, x)

  ϵ0 = @. eos(InternalEnergy, Density, ρ0, Pressure, p0)

  @unpack D, S, tau = get_dynamic_variables(cache)
  @unpack p         = get_static_variables(cache)
  @. p   = p0
  @. D   = Formulae.D(ρ0, v0, ϵ0, p0)
  @. S   = Formulae.S(ρ0, v0, ϵ0, p0)
  @. tau = Formulae.τ(ρ0, v0, ϵ0, p0)

  broadcast_volume!(cons2prim, P.equation, cache)

  return
end


function initialdata_equation_of_state(::Val{type}, eos, x) where type
  error("initialdata_equation_of_state: Unknown initial data type '$type'")
end


function initialdata_equation_of_state(::Val{:smooth}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)
  @unpack gamma = eos

  # from Bugner, PhD thesis 2018
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert gamma == 5/3
  ρ0 = @. 1.0 + 0.2 * sin(2.0 * π * x)
  v0 = @. 0.2 * x^0
  p0 = @. 1.0 * x^0

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_relativistic_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # inspired by Bugner, PhD thesis 2018; Marti, Mueller, JCP 1996
  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  : 1.0   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   : 0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 13.33 : 1e-7  for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_relativistic_shocktube_2}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # inspired by Bugner, PhD thesis 2018; Marti, Mueller, JCP 1996
  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 1.0    : 1.0   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0    : 0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 1000.0 : 0.01  for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_shocktube_pressure_jump}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)
  γ          = eos.gamma

  # selfmade, less extreme case than periodic_relativistic_shocktube with only a jump
  # in pressure
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert γ == 5/3
  x0 = 0.5
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  :  10.0 for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   :  0.0  for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 13.33 :  0.1  for xi in x ]
  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_shocktube_less_extreme}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # selfmade, less extreme case than periodic_relativistic_shocktube
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert gamma == 5/3
  x0 = 0.5
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  :  0.5   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   :  0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 8.0   :  1.5 for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_sod_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 1.0  :  0.125  for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0  :  0.0    for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 1.0  :  0.1    for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:sod_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 0.0
  ρ0 = [ (xi < x0) ? 1.0  :  0.125  for xi in x ]
  v0 = [ (xi < x0) ? 0.0  :  0.0    for xi in x ]
  p0 = [ (xi < x0) ? 1.0  :  0.1    for xi in x ]

  (ρ0, v0, p0)
end


#######################################################################
#                   Initial data entropy variables                    #
#######################################################################


function initialdata_entropy_variables!(env, P::Project, prms)
  @TODO
  @unpack cache = env
  @unpack t, tm1 = get_global_variables(cache)
  @unpack E, Em1, flx_E, flx_Em1 = get_static_variables(cache)
  tm1 .= 0.0
  t   .= 0.0
  compute_entropy_variables!(P, P.equation.eos)
  @. Em1 = E
  @. flx_Em1 = flx_E
end


#######################################################################
#                          Initial data HRSC                          #
#######################################################################


function initialdata_hrsc!(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, prms) end


function initialdata_hrsc!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mu = get_cell_variables(env.cache)
  @. mu = 0.0
end


function initialdata_hrsc!(env, P::Project, hrsc::HRSC.SmoothedArtificialViscosity, prms)
  initialdata_hrsc!(env, P, hrsc.av, prms)
  @unpack smoothed_mu = get_static_variables(env.cache)
  @. smoothed_mu = 0.0
end


#######################################################################
#                          Initial data TCI                           #
#######################################################################


function initialdata_tci!(env, P::Project, tci::Nothing, prms) end


function initialdata_tci!(env, P::Project, tci::TCI.EntropyProduction, prms)
  @unpack cache = env
  @unpack EP = get_static_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @. EP = 0.0
  @. EPjump_lhs = 0.0
  @. EPjump_rhs = 0.0
end


function initialdata_tci!(env, P::Project, tci::TCI.AbstractTCI, prms)
  @unpack flag = get_cell_variables(env.cache)
  @. flag = 0.0
end
