function Project(env::Environment, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = make_Equation(eos, prms["SRHD"])
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])
  rsolver = ApproxRiemannSolver(flux, speed, equation)
  ldg_rsolver, av_rsolver = if hrsc isa AbstractArtificialViscosity
    ApproxRiemannSolver(ldg_flux, ldg_speed, equation),
    ApproxRiemannSolver(av_flux, av_flux, equation)
  else
    nothing, nothing
  end

  P = Project(equation, rsolver, hrsc, tci, ldg_rsolver, av_rsolver)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(cache, P)
  register_variables!(cache, rsolver)
  !isnothing(ldg_rsolver) && register_variables!(cache, ldg_rsolver)
  !isnothing(av_rsolver) && register_variables!(cache, av_rsolver, dont_register=true)
  display(cache)

  # setup boundary conditions
  bdryconds = make_BoundaryConditions(env, equation, rsolver, prms["SRHD"])
  ldg_bdryconds, av_bdryconds = if hrsc isa AbstractArtificialViscosity
    make_BoundaryConditions_LDG(env, equation, ldg_rsolver, prms["SRHD"]),
    make_BoundaryConditions_LDG(env, equation, av_rsolver, prms["SRHD"])
  else
    nothing, nothing
  end

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup initial data
  initialdata!(env, P, prms["SRHD"])

  # setup callbacks
  projectcb = make_callback(env, P, bdryconds)

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(cache, P::Project)
  register_variables!(cache,
    dynamic_variablenames = (:D, :S, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_S, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D, :flx_S, :flx_tau,  # conservatives' fluxes
                             :rho, :v, :p, :eps,        # primitives
                             :max_v, :x,                # maximum charactersitic speed, grid
                             :E, :Em1,                  # entropy
                             :flx_E, :flx_Em1),         # entropy flux
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  _register_variables!(cache, P.hrsc)
  _register_variables!(cache, P.tci)
  @unpack x = get_static_variables(cache)
  # @. x = env.mesh.x[:]
end


function _register_variables!(cache, tci_or_hrsc::Nothing) end


_register_variables!(cache, hrsc::AbstractHRSC) = 
  error("_register_variables: Not implemented for hrsc of type '$(typeof(hrsc))'")

_register_variables!(cache, hrsc::HRSC.AbstractReconstruction) = nothing


function _register_variables!(cache, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(cache,
    static_variablenames  = (:ldg_D, :ldg_S, :ldg_tau,              # local DG variable
                             :flx_ldg_D, :flx_ldg_S, :flx_ldg_tau), # local DG flux
    cell_variablenames    = (:mu,), # 'one viscosity to rule them all'
    )
end


function _register_variables!(cache, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(cache, hrsc.av)
  register_variables!(cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(cache, tci::TCI.AbstractTCI)
  register_variables!(cache, cell_variablenames = (:flag,:D_flag,:S_flag,:tau_flag))
end


function _register_variables!(cache, tci::TCI.EntropyProduction)
  register_variables!(cache,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end
