module TCI


using dg1d
using LinearAlgebra
import Parameters: @unpack


export make_Indicator,
       Indicator,
       compute_indicator!,
       make_callback


include("interface.jl")
# implementations
include("minmod.jl")
include("threshold.jl")
include("entropyproduction.jl")
include("diffusion.jl")
include("modaldecayaverage.jl")
include("modaldecayhighest.jl")


@parameters TCI begin

  """
  Available options:
  - `none` ... disabled
  - `diffusion` ... based on gradient of velocity field
  - `mda` ... based on average of modal decay of solution
  - `mdh` ... based on modal decay of highest order coefficient of solution
  - `minmod` ... based on minmod limiter to detect undesired oscillations
  - `threshold` ... flag if threshold is exceeded
  - `entropy_production` ... flag if entropy production is exceeded
  """
  method = "none"
  @check method in [ "none", "diffusion", "mda", "mdh", "minmod", "threshold",
                     "entropy_production" ]

  "List of variable names to monitor. If more than one is given `combine_method`
  controls how to add the TCI results up."
  variables = String[]
  @check length(variables) > 0

  """
  This controls on how to combine TCIs from different variables. Available options:
  - `max`
  - `min`
  - `average`
  - `sum`
  """
  combine_method = "max"
  @check combine_method in [ "max", "min", "average", "sum" ]

  "Flag if diffusion exceeds this value"
  diffusion_max = 1.0
  @check diffusion_max > 0

  "Flag if modal decay rate falls below this value"
  mda_threshold_min = 1.0
  @check mda_threshold_min > 0

  "Don't flag if modal decay rate exceeds this value"
  mda_threshold_max = 3.0
  @check mda_threshold_max > mda_threshold_min

  "Flag if modal decay rate falls below this value"
  mdh_threshold_min = 1.0
  @check mdh_threshold_min > 0

  "Don't flag if modal decay rate exceeds this value"
  mdh_threshold_max = 3.0
  @check mdh_threshold_max > mdh_threshold_min

  "TODO"
  minmod_threshold = 1e-5
  @check minmod_threshold > 0

  "Flag if monitor exceeds this value"
  threshold = 1e-5
  @check threshold > 0

  "Flag if entropy production exceeds this value"
  entropy_production_max = 1.0
  @check entropy_production_max > 0

end


end
