"""
    main(parfile)

Run program specified by a `parfile` in TOML format.
"""
function main(parfile)

  # normalize parfile name
  parfile = abspath(parfile)

  project_name, prms = load_parameters(parfile)

  # TODO Use PrettyTables.jl for summary
  # TODO Dump parameters also to file
  # TODO Dump versioninfo() to some file
  # TODO Dump Pkg.status() to some file
  # TODO Dump `git status` to some file
  # TODO Dump gethostname() and other system info
  # summary = summarize(prms)
  # println(summary)

  # backup any previous output dir and setup a fresh one
  # TODO Rename this to make_outputdir
  outputdir = mk_outputdir(parfile)
  cp(parfile, joinpath(outputdir, basename(parfile)))

  # Setup the environment
  mesh = Mesh(; N=prms["Mesh"]["n"], K=prms["Mesh"]["k"], xrange=prms["Mesh"]["xrange"],
              basis=prms["Mesh"]["basis"], CFL=prms["Mesh"]["cfl"])
  cache = Cache(mesh)
  callbacks = CallbackSet()
  callbacks_substeps = CallbackSet()
  env = Environment(mesh, cache, callbacks, callbacks_substeps)

  # Setup the project. This involves
  # - registering variables in cache
  # - setting up initial data
  # - setting up boundary conditions
  # - setting up callbacks
  # Need to use Val(project_name) for dispatch here, because project_name is a runtime variable.
  Project = project_type(Val(project_name))
  # TODO Till I sorted out the issue with where to place bdry condition types, I have to
  # return them here and forward them below
  project, bdryconds = Project(env, prms)

  # wrap project interface methods for useage with time stepper
  function wrapped_rhs!(state_du, state_u, t)
    wrap_dynamic_variables!(env.cache, state_u)
    wrap_rhs_variables!(env.cache, state_du)
    # provided by project interface
    rhs!(env, project, bdryconds)
  end
  function wrapped_timestep!(state_u, t, n)
    wrap_dynamic_variables!(env.cache, state_u)
    # provided by project interface
    timestep!(env, project)
  end

  # Setup project independent callbacks
  # For the time being, we let the Projects freely push callbacks into env.callbacks.
  # This should make no difference as long as we 1) don't run in parallel and 2) don't allow
  # multiple projects. Might want to move that logic later from Project to here.
  # TODO Add an AliveCallback that manages a 'status' file in the output folder
  # TODO Add priority ordering to CallbackSet and callbacks.
  # TODO Drop unused callbacks from callbackset upon creation

  # Add some 'global' callbacks

  savecb = SaveCallback(Symbol.(prms["Output"]["variables"]), cache, mesh,
                        joinpath(outputdir, "output.h5"),
                        CallbackTiming(every_dt=prms["Output"]["every_dt"],
                                       every_iteration=prms["Output"]["every_iteration"]))

  aligned_ts = prms["Output"]["aligned_ts"]
  timealigned_savecb, wrapped_timestep! = if length(aligned_ts) > 0
    atscb = TimeAlignedSaveCallback(aligned_ts, wrapped_timestep!, savecb)
    atscb, atscb.aligned_timestep_fn
  else
    nothing, wrapped_timestep!
  end

  plotfn(u,t) = plot(env, project)
  plotcb = PlotCallback(plotfn, CallbackTiming(every_dt=prms["Plot"]["every_dt"],
              every_iteration=prms["Plot"]["every_iteration"]), pause=prms["Plot"]["pause"])

  push!(callbacks, savecb)
  push!(callbacks, timealigned_savecb)
  push!(callbacks, plotcb)

  # TODO Setup a logger, maybe want to setup even earlier
  # TODO summarize current setup once and log to screen
  # println(summarize(env.mesh))
  # println(summarize(env.cache))
  # println(summarize(env.callbacks))
  # println(summarize(env.callbacks_substep))
  # println(summarize(project))

  # plot parameter list
  # TODO Dumb this to a file
  for (k,v) in prms
    table = vcat([k ""], hcat(collect(keys(v)), collect(values(v))))
    pretty_table(
        table;
        body_hlines = [1],
        body_hlines_format = Tuple('─' for _ = 1:4),
        alignment = [ :l, :l ],
        noheader = true,
        tf = tf_borderless,
        columns_width = 30,
        linebreaks = true,
        autowrap = true,
        crop = :none
    )
    println()
  end

  # Let's go!!!
  evolve!(env, wrapped_rhs!, wrapped_timestep!, prms)

  return
end


function load_parameters(parfile)

  # read parameters
  if !isfile(parfile)
    error("Cannot locate parameter file '$parfile'")
  end

  user_prms = TOML.parsefile(parfile)
  # switch to a flat dict where keys of nested dicts are concatenated with dots
  user_prms = Dict(string(k) => flatten(v) for (k,v) in pairs(user_prms))

  ### sanitize parameter sections
  # supported top-level headers that aren't user projects
  default_projects = [ :Mesh, :Evolution, :Plot, :Output, :TCI, :HRSC, :EquationOfState ]

  for (k,v) in user_prms
    if !(v isa AbstractDict)
      error("Invalid parameter file: No top-level parameters allowed, found '$k = $v'")
    end
  end

  # find a parameter section for which somebody registered a project with that name
  all_projects = Symbol.(keys(user_prms))
  maybe_user_projects = filter(all_projects) do k
    return !(k in default_projects)
  end
  registered_projects = filter(isregistered ∘ Symbol, maybe_user_projects)
  not_registered_projects = setdiff(maybe_user_projects, registered_projects)
  if length(not_registered_projects) > 0
    error("Invalid parameter file: Found unknown top-level sections: $(not_registered_projects)")
  end
  if length(registered_projects) == 0
    error("Invalid parameter file: Missing a parameter section for a project, don't know which project to run ...")
  end
  if length(registered_projects) > 1
    error("Invalid parameter file: Only one project parameter section allowed, found multiple: $(registered_projects)")
  end

  # we found a user project
  project_name = Symbol(first(registered_projects))

  # load default parameters
  prms = Dict(string(s) => parameters(Val(s)) for s in default_projects)
  prms[string(project_name)] = parameters(project_name)
  # TODO parameters(Val) does not guarantee that dicts returned from parameter are flat.
  prms = Dict(k => flatten(v) for (k,v) in pairs(prms))

  # merge defaults with with user parameters
  for (sec, u_prms) in pairs(user_prms)
    def_prms = prms[sec]
    for (k,v) in pairs(u_prms)
      if !haskey(def_prms, k)
        error("Invalid parameter file: Unsupported parameter found in section '$sec': '$k = $v'. Keep your parameter files clean!")
      end
      def_prms[k] = v
    end
  end

  # let's keep things organized
  prms = OrderedDict( k => OrderedDict(v) for (k,v) in pairs(prms) )
  # sort sections
  sort!(prms)
  # sort each section
  foreach(p -> sort!(p), values(prms))

  return project_name, prms
end


function evolve!(env, rhs_fn, timestep_fn, prms)

  @unpack tend, method = prms["Evolution"]
  @unpack cfl = prms["Mesh"]
  @unpack cache, callbacks, callbacks_substep = env

  # sanity checks
  if tend <= 0
    error("TimeStepper: tend must be positive, found 'tend = $tend'")
  end
  if cfl < 0
    error("TimeStepper: cfl must be positive, found 'cfl = $cfl'")
  end
  methods = [ "midpt", "rk4", "rk4_twothirds", "rk4_ralston", "ssprk3", "lserk4", "rkf10" ]
  if !(method in methods)
    error("TimeStepper: Unknown time stepping method '$method' requested, available methods: '$(methods...)'")
  end

  # TODO This should be hidden behind an interface
  all_vars = values(cache.dynamic_variables.datadict)
  u0 = vcat(all_vars...)

  # Let's go!!!
  evolve_ERK(rhs_fn, u0, timestep_fn, (0.0, tend), Symbol(method),
             callback_fullstep = callbacks)#, callback_substep = callbacks_substep)

end
