"""
    @parameters name ex

Macro to define parameters and their default values for a symbol `name`.
If invoked with `MyProject` this generates the following methods for the `dg1d` project interface:
- `dg1d.parameters(::Val{:MyProject})`
- `dg1d.check_parameters(::Val{:MyProject}, prms)`

---

# Example

```julia
@parameters MyProject begin

  "some variable"
  x = 1.0
  @check x > 0.0

  "another variable"
  y = "String"

end
```
"""
macro parameters(name, expr)
  return esc(_generate_parameters(name, expr))
end


macro check(ex)
  return ex
end


function _generate_parameters(name, expr)

  if expr.head !== :block
    error("@parameters: Second argument must be a begin ... end block containing docstrings and parameter defintions!")
  end

  pars = expr.args
  pars = filter(p -> !(p isa LineNumberNode), pars)

  docstrings = String[]
  kvexprs = Expr[]
  checks = Vector{Expr}[]
  i_definition = 0
  for p in pars

    # if p isa Expr
    #   filter!(pp -> !(pp isa LineNumberNode), p.args)
    # end

    is_definition = (p isa Expr) && p.head === :macrocall &&
       (p.args[1] === Symbol("@doc") || p.args[1] isa GlobalRef) && length(p.args) == 4
    is_check = (p isa Expr) && p.head === :macrocall && p.args[1] === Symbol("@check") &&
      length(p.args) == 3

    is_valid_block = is_definition || is_check
    if !is_valid_block
      error("""
            @parameters: Invalid parameter definition.
            Expected either a block with a docstring, parameter name and default value
            or an argument check, e.g.
               \"this is a docstring for variable x\"
               x = 1.0
               @check x > 1.0
            Instead found
               $(string(p))
            """)
    end

    if is_definition
      # parse a parameter definition block

      i_definition += 1

      # extract docstring, mandatory
      d = p.args[3]
      if d isa Expr && d.head === :macrocall && d.args[1] === Symbol("@raw_str")
        d = d.args[3]
      end

      # extract variable name and default value, mandatory
      kv = p.args[4]
      if !(kv isa Expr)
        error("""
              @parameters: Invalid parameter definition. Missing default value for
                 $(string(kv))
              """)

      end
      if kv.head !== :(=)
        error("""
              @parameters: Invalid parameter definition. Expected pair of parameter name and default value, e.g.
                 x = 1.0
              Instead found
                 $(string(kv))
              """)
      end

      k, v = kv.args[1], kv.args[2]
      if !(k isa Symbol)
        error("""
              @parameters: Invalid parameter name: '$k'"
              """)
      end

      push!(docstrings, d)
      push!(kvexprs, kv)
      push!(checks, Expr[])
    end

    # parse a check, they are optional
    if is_check

      # we already verified above that p is a @check block
      # we also stripped any LineNumberNodes there
      ck = p.args[3]

      if i_definition == 0
        error("@parameters: Invalid definition. Found @check before any parameter definition: $ck")
      end

      # macro expansion of @check is delayed till check_parameters definition below
      push!(last(checks), ck)
    end

  end

  keys = [ kv.args[1] for kv in kvexprs ]
  for i = 1:length(keys)-1
    refk = keys[i]
    foreach(keys[i+1:end]) do k
      if refk === k
        error("""
              @parameters: Duplicated parameters found: '$k'
              """)
      end
    end
  end

  # generate the docs
  io = IOBuffer()
  println(io, name)
  for (d, kv) in zip(docstrings, kvexprs)
    println(io, "- `", kv, "`\n    ", replace(d, '\n' => "\n  "))
  end
  docs = String(take!(io))

  # compare key = value to "key" => value for Dict constructor
  pairs = [ Expr(:call, :(=>), string(kv.args[1]), kv.args[2]) for kv in kvexprs ]

  checkscode = Expr[]
  for cks in checks
    evalchecks = [ quote
                     result = try
                       $ck
                     catch
                       error("check_parameters for $($(string(name))): Failed to evaluate check '$($(string(ck)))'")
                     end
                     if !(result isa Bool)
                       error("check_parameters for $($(string(name))): Invalid check found '$($(string(ck)))'. Expected to return a Bool, but found $(typeof(result))")
                     end
                     if result == false
                       error("check_parameters for $($(string(name))): Check failed: '$($(string(ck)))'")
                     end
                   end
                   for ck in cks ]
    code = quote $(evalchecks...) end
    push!(checkscode, code)
  end
  unpack_prms = [ quote $k = prms["$($("$k"))"] end for k in keys ]

  code = quote
    @doc $docs dg1d.parameters(::Val{$(QuoteNode(name))}) = Dict($(pairs...))
    function dg1d.check_parameters(::Val{$(QuoteNode(name))}, prms)
      $(unpack_prms...)
      $(checkscode...)
    end
  end

  return code
end


#######################################################################
#                        Parameter definitions                        #
#######################################################################


@parameters Mesh begin

  "Number of cells"
  k = 10
  @check k > 0

  "Physical domain extensions in x direction"
  xrange = [ -1.0, 1.0 ]
  @check length(xrange) == 2
  @check xrange[2] - xrange[1] > 0.0

  # TODO Move to Evolution
  "Courant-Friedrichs-Lewy factor"
  cfl = 0.5
  @check cfl > 0.0

  "Polynomial order in each grid"
  n = 5
  @check n > 0

  """
  Nodal basis used for the polynomial approximation. Available options:
  - `lgl_lumped` ... mass-lumped Legendre-Gauss-Lobatto grid
  - `lgl`        ... Legendre-Gauss-Lobatto grid
  - `glgl`       ... Generalized Legendre-Gauss-Lobatto grid
  """
  basis = "lgl_lumped"
  @check basis in [ "lgl", "glgl", "lgl_lumped" ]

end


@parameters Evolution begin

  "Final time"
  tend = 10.0
  @check tend > 0.0

  """
  Time stepping method. Available options
  - `midpt`         ... 2nd order explicit Runge Kutta method -- midpoint formula
  - `rk4`           ... 4th order explicit Runge Kutta method
  - `rk4_twothirds` ... 4th order explicit Runge Kutta method -- two thirds version
  - `rk4_ralston`   ... 4th order explicit Runge Kutta method -- Ralston's method,
  -                   designed for minimal truncation error
  - `ssprk3`        ... 3rd order strong stability preserving Runge Kutta method
  - `lserk4`        ... 4th order low storage explicit Runge Kutta method
  - `rkf10`         ... 10th order Runge Kutta Feagin method
  """
  method = "lserk4"
  @check method in [ "midpt", "lserk4", "rk4", "rk4_twothirds", "rk_ralston",
                     "ssprk3", "rkf10" ]

end


@parameters Plot begin

  "Call plot callback after every_iteration full time steps"
  every_iteration = 0
  @check every_iteration >= 0

  "Call plot callback after every_dt simulation time"
  every_dt = 0.0
  @check every_dt >= 0

  """
  If true, pauses the simulation to display the plot.
  If set to a positive number, the simulations is paused only for that
  number of seconds (walltime).
  """
  pause = false
  @check pause isa Bool || pause > 0.0

  "List of variables that should be plotted"
  variables = String[]

end


@parameters Output begin

  "Output data after every_iteration full time steps"
  every_iteration = 0
  @check every_iteration >= 0

  "Output data after every_dt simulation time"
  every_dt = 0.0
  @check every_dt >= 0

  "List of variables that should be outputted"
  variables = String[]

  "Filename for .h5 output"
  filename = ""
  @check length(filename) > 0

  "List of times at which data should be outputted"
  aligned_ts = Float64[]
  @check all(t -> t > 0, aligned_ts)
  @check issorted(aligned_ts)

end
