module IntegrationTests

export runtests

using dg1d
using Printf
using PrettyTables
using LinearAlgebra


const REFDIR = abspath(joinpath(@__DIR__, "..", "refs"))
const TESTDIR = abspath(joinpath(@__DIR__, "..", "tests"))


"""
    discover_output_dirs(rootdir)

A directory named `testdir` is an output directory if
- it contains a parameter file name `testdir.toml`
- it contains an output file name `output.h5`
"""
function discover_output_dirs(rootdir)

    _, dirs, _ = first(walkdir(rootdir))

    refdirs, refparfiles, refoutputs = String[], String[], String[]
    for d in dirs
        parfile = "$d.toml"
        output = "output.h5"
        !isfile(joinpath(rootdir, d, parfile)) && continue
        !isfile(joinpath(rootdir, d, output)) && continue
        push!(refdirs, d)
        push!(refparfiles, parfile)
        push!(refoutputs, output)
    end

    return refdirs, refparfiles, refoutputs
end


function prepare_tests(refdirs, refpars)

    # clean up previous tests
    if isdir(TESTDIR)
        rm(TESTDIR, recursive=true, force=true)
    end
    mkdir(TESTDIR)

    pars = String[]
    for (refdir, refpar) in zip(refdirs, refpars)
        rpar = joinpath(REFDIR, refdir, refpar)
        par = joinpath(TESTDIR, refpar)
        cp(rpar, par)
        push!(pars, par)
    end

    return pars
end


function run_pars(pars)

    success = Bool[]
    dirs = String[]
    for p in pars
        @info "Running parameter file '$p' ..."
        s = try
            dg1d.main(p)
            true
        catch
            false
        end
        push!(success, s)
    end

    return success
end


function compare_outputs(io, refoutput, output,
        variables, reltol=1e-8, abstol=1e-10)::Bool

    comparsion_results = Any[]

    rout = dg1d.Output(refoutput)
    out = dg1d.Output(output)

    ref_meshlayout = size(rout.mesh)
    meshlayout = size(out.mesh)
    if !all(ref_meshlayout .== meshlayout)
        println(io, """
                    FAIL: Mismatch in mesh size
                    reference $ref_meshlayout
                    test      $meshlayout
                    """)
        return false
    end

    refts = rout.times
    ts = out.times
    if length(refts) != length(ts)
        println(io, """
                    FAIL: Mismatch in number of time steps for which outputs were recorded
                    reference $(@sprintf("%10d",length(refts))) : $refts
                    test      $(@sprintf("%10d", length(ts))) : $ts
                    """)
        return false
    end

    if any(zip(refts, ts)) do (rti, ti)
            !isapprox(rti, ti)
        end

        println(io, "FAIL: Mismatch in output times:\n")
        data = hcat(refts, ts)
        pretty_table(io, data;
            header = [ :reference, :test ],
            tf = tf_simple,
            formatters = ft_printf("% 10.10e"),
            crop = :none)
        return false
    end

    success = true
    for var in variables
        if !dg1d.has_variable(rout, var)
            println(io, "FAIL: Missing variable '$var' in reference output")
            success = false
            continue
        end
        if !dg1d.has_variable(out, var)
            println(io, "FAIL: Missing variable '$var' in test output")
            success = false
            continue
        end

        for i in 1:length(ts)
            refvi = read_variable(rout, var, i)
            vi = read_variable(out, var, i)

            if !all(zip(refvi, vi)) do (rv, v)
                    isapprox(rv, v, rtol=reltol, atol=abstol)
                end
                println(io, "FAIL: Mismatch in variable '$var' @ (t,i) = ($(@sprintf("%10.10e",ts[i])),$i), norm(ref-test,1) = $(norm(refvi.-vi,1))")
                success = false
            end
        end
    end

    return success
end


"""
    runtests(tests::Vector{String})

Run integration tests for which there is reference data
in `refs/` and a configuration entry in `refs/testconfig.toml`.
"""
function runtests(tests::Vector{String}=String[])

    allrefdirs, allrefpars, allrefoutputs = discover_output_dirs(REFDIR)

    # load and check test config
    filename = normpath(joinpath(REFDIR, "testconfig.toml"))
    testconfig = TOML.parsefile(filename)

    default_reltol = 1e-8
    default_abstol = 1e-10
    default_reltol = get!(testconfig, "default_reltol", default_reltol)
    default_abstol = get!(testconfig, "default_abstol", default_abstol)
    if default_reltol < 0
        error("Invalid testconfig.toml: 'default_reltol' must be positive, found 'default_reltol = $default_reltol'")
    end
    if default_abstol < 0
        error("Invalid testconfig.toml: 'default_abstol' must be positive, found 'default_abstol = $default_abstol'")
    end

    # drop default tolerance paraemeters
    testconfig = filter(testconfig) do (k,_)
        !(k in ["default_reltol", "default_abstol"])
    end
    display(testconfig)

    for (registered_test, pars) in testconfig
        if !(pars isa AbstractDict)
            error("Invalid testconfig.toml: Found unsupported parameter '$registered_test'")
        end
        # if no vars are given, we compare all available ones
        vars = get!(pars, "variables", String[])
        reltol = get!(pars, "reltol", default_reltol)
        abstol = get!(pars, "abstol", default_abstol)
        if default_reltol < 0
            error("Invalid testconfig.toml: '$registered_test.reltol' must be positive, found '$registered_test.reltol = $reltol'")
        end
        if default_abstol < 0
            error("Invalid testconfig.toml: '$registered_test.abstol' must be positive, found '$registered_test.abstol = $abstol'")
        end
        if !(registered_test in allrefdirs)
            error("Invalid testconfig.toml: Missing reference data for test '$registered_test'")
        end
    end

    if length(tests) == 0
        # run all available tests
        tests = allrefdirs
    else
        foreach(tests) do t
            if !(t in allrefdirs)
                error("Unknown test '$t' request. Make sure there is a folder named '$t` in `refs/` and a configuration entry in `refs/testconfig.toml`")
            end
        end
    end

    # gather dirs, parfiles and outputs for requested tests
    idxs = [ findfirst(rd -> rd == t, allrefdirs) for t in tests ]
    refdirs = allrefdirs[idxs]
    refpars = allrefpars[idxs]
    refoutputs = allrefoutputs[idxs]

    # prepare tests folder and copy parfiles
    pars = prepare_tests(refdirs, refpars)
    dirs = [ first(splitext(p)) for p in pars ]
    outputs = [ "output.h5" for _ in pars ]

    # run the tests
    run_success = run_pars(pars)

    # run comparsions on test outputs
    logs = IOBuffer()
    n_header = 80
    comparsion_success = Union{Bool,Missing}[]
    for i in 1:length(tests)

        test = tests[i]
        println(logs, repeat('+', n_header), "\n\nTest: $test\n")

        if !run_success[i]
            println(logs, "FAIL: Failed to run test!")
            push!(comparsion_success, missing)
            println(logs, "\nResult: FAILED\n")
            continue
        end

        refout = joinpath(REFDIR, refdirs[i], refoutputs[i])
        out    = joinpath(TESTDIR, dirs[i], outputs[i])
        @unpack variables, reltol, abstol = testconfig[test]
        comp_success = compare_outputs(logs, refout, out, variables, reltol, abstol)
        push!(comparsion_success, comp_success)

        result = run_success[i] && comp_success

        println(logs, result ? "" : "\n", "Result: ", result ? "PASSED" : "FAILED", "\n")
    end

    # print overall summary
    n_runs = length(run_success)
    n_runs_succeeded = count(run_success)
    n_runs_failed = n_runs - n_runs_succeeded
    pct_runs_succeeded = n_runs_succeeded / n_runs * 100

    comparsions_done = filter(!ismissing, comparsion_success)
    n_compars = length(comparsions_done)
    n_compars_succeeded = count(comparsions_done)
    n_compars_failed = n_compars - n_compars_succeeded
    pct_compars_succeeded = n_compars_succeeded / n_compars * 100

    final_result = (n_runs_succeeded == n_runs) && (n_compars_succeeded == n_compars)

    println(logs, repeat('=', n_header))
    println(logs, repeat('+', n_header))
    println(logs, repeat('=', n_header))
    println(logs, """

Summary

Tests succeded:         $(@sprintf("%3d/%-3d = %3.1f%%", n_runs_succeeded, n_runs, pct_runs_succeeded))
Comparsions succeded:   $(@sprintf("%3d/%-3d = %3.1f%%", n_compars_succeeded, n_compars, pct_compars_succeeded))
Final result:           $(final_result ? "PASSED" : "FAILED")
    """)

    logfilename = joinpath(TESTDIR, "tests.log")
    logstr = String(take!(logs))
    println(logstr)
    write(logfilename, logstr)


    return final_result
end

end
