@testset "Cache" begin


@testset "Constructor and getters" begin

  # constructor
  mesh = Mesh()
  C = dg1d.Cache(mesh)

  # getters
  @test n_variables(get_dynamic_variables(C)) == 0
  @test n_variables(get_rhs_variables(C))     == 0
  @test n_variables(get_static_variables(C))  == 0
  @test n_variables(get_cell_variables(C))    == 0
  @test n_variables(get_bdry_variables(C))    == 0
  @test n_variables(get_global_variables(C))  == 0

end


@testset "Getters" begin

  mesh = Mesh()
  C = dg1d.Cache(mesh)
  dynamic_variablenames = (:u, :v)
  rhs_variablenames     = (:rhs_u, :rhs_v)
  dg1d.register_variables!(C;
                           dynamic_variablenames = dynamic_variablenames,
                           rhs_variablenames     = rhs_variablenames)

  groupnames = sort(variablegroups(C))
  ref_groupnames = sort([ :dynamic_variables, :rhs_variables, :static_variables,
                          :cell_variables, :bdry_variables, :global_variables ])
  @test all(groupnames == ref_groupnames)

  varnames = sort(variablenames(C))
  ref_varnames = sort(vcat(dynamic_variablenames..., rhs_variablenames...))
  @test all(varnames .== ref_varnames)

  dyn_varnames = sort(variablenames(C, :dynamic_variables))
  ref_dyn_varnames = sort([ d for d in dynamic_variablenames])
  @test all(dyn_varnames .== ref_dyn_varnames)
  rhs_varnames = sort(variablenames(C, :rhs_variables))
  ref_rhs_varnames = sort([ d for d in rhs_variablenames])
  @test all(rhs_varnames .== ref_rhs_varnames)

  @unpack u, v = get_dynamic_variables(C)
  @test u === get_variable(C, :u)
  @test v === get_variable(C, :v)

  @unpack rhs_u, rhs_v = get_rhs_variables(C)
  @test rhs_u === get_variable(C, :rhs_u)
  @test rhs_v === get_variable(C, :rhs_v)

  @test u     === get_variable(C, :u, :dynamic_variables)
  @test v     === get_variable(C, :v, :dynamic_variables)
  @test rhs_u === get_variable(C, :rhs_u, :rhs_variables)
  @test rhs_v === get_variable(C, :rhs_v, :rhs_variables)

  # variable names must be unique!
  C = dg1d.Cache(mesh)
  @test_throws ErrorException dg1d.register_variables!(C;
                                                       dynamic_variablenames = (:u,),
                                                       rhs_variablenames     = (:u,))

end


@testset "Registering" begin

  mesh = Mesh()
  C = dg1d.Cache(mesh)

  dynamic_variablenames = (:u, :v)
  rhs_variablenames     = (:rhs_u, :rhs_v)
  static_variablenames  = (:f_u, :f_v)
  cell_variablenames    = (:max_speed,)
  bdry_variablenames    = (:nf_u_lhs, :nf_u_rhs, :nf_v_lhs, :nf_v_rhs)
  kwargs = (; dynamic_variablenames, rhs_variablenames, static_variablenames,
            cell_variablenames, bdry_variablenames)

  # register variables
  dg1d.register_variables!(C; kwargs...)

  # getters
  names = dg1d.names(get_dynamic_variables(C))
  @test tuple(names...) == dynamic_variablenames
  names = dg1d.names(get_rhs_variables(C))
  @test tuple(names...) == rhs_variablenames
  names = dg1d.names(get_static_variables(C))
  @test tuple(names...) == static_variablenames
  names = dg1d.names(get_cell_variables(C))
  @test tuple(names...) == cell_variablenames
  names = dg1d.names(get_bdry_variables(C))
  @test tuple(names...) == bdry_variablenames

end


@testset "Wrapping" begin

  mesh = Mesh()

  # register some variables
  dynamic_variablenames = (:var1, :var2)
  rhs_variablenames     = (:rhs_var1, :rhs_var2)
  mesh = Mesh()
  C = dg1d.Cache(mesh)
  register_variables!(C; dynamic_variablenames = dynamic_variablenames,
                         rhs_variablenames     = rhs_variablenames)

  # allocate some data independent of cache
  NN      = length(mesh)
  n_vars  = n_variables(get_dynamic_variables(C))
  data_u  = randn(n_vars * NN)
  data_du = randn(n_vars * NN)
  # and wrap the cache around it
  wrap_dynamic_variables!(C, data_u)
  wrap_rhs_variables!(C, data_du)

  @unpack var1, var2         = get_dynamic_variables(C)
  @unpack rhs_var1, rhs_var2 = get_rhs_variables(C)
  @test all(var1     .== data_u[1:NN])
  @test all(var2     .== data_u[NN+1:2*NN])
  @test all(rhs_var1 .== data_du[1:NN])
  @test all(rhs_var2 .== data_du[NN+1:2*NN])

  # this repeats the above test but with order of dynamic variables reversed when registering
  # hence, the order in which variables are registered matters!
  dynamic_variablenames = (:var2, :var1)
  rhs_variablenames     = (:rhs_var1, :rhs_var2)
  mesh = Mesh()
  C = dg1d.Cache(mesh)
  register_variables!(C; dynamic_variablenames = dynamic_variablenames,
                         rhs_variablenames     = rhs_variablenames)

  NN      = length(mesh)
  n_vars  = n_variables(get_dynamic_variables(C))
  data_u  = randn(n_vars * NN)
  data_du = randn(n_vars * NN)
  wrap_dynamic_variables!(C, data_u)
  wrap_rhs_variables!(C, data_du)

  @unpack var1, var2           = get_dynamic_variables(C)
  @unpack rhs_var1, rhs_var2 = get_rhs_variables(C)
  # revsered check for var1, var2
  @test all(var2     .== data_u[1:NN])
  @test all(var1     .== data_u[NN+1:2*NN])
  @test all(rhs_var1 .== data_du[1:NN])
  @test all(rhs_var2 .== data_du[NN+1:2*NN])

end


@testset "Printing" begin

  mesh = Mesh()

  # register some variables
  dynamic_variablenames = (:var1, :var2)
  rhs_variablenames     = (:rhs_var1, :rhs_var2)
  bdry_variablenames    = (:a, :b, :c)
  cell_variablenames    = (:cell_speed, :cell_mean, :cell_avg)
  static_variablenames  = (:entropy, :entropy_flux)
  global_variablenames  = (:t,)
  mesh = Mesh()
  C = dg1d.Cache(mesh)
  register_variables!(C; dynamic_variablenames = dynamic_variablenames,
                         rhs_variablenames     = rhs_variablenames,
                         bdry_variablenames    = bdry_variablenames,
                         cell_variablenames    = cell_variablenames,
                         static_variablenames  = static_variablenames,
                         global_variablenames  = global_variablenames)
  # should print a table with the above variable names in a separate column for each variable
  display(C)


  # test with only some registered variables
  C = dg1d.Cache(mesh)
  register_variables!(C; dynamic_variablenames = dynamic_variablenames,
                         bdry_variablenames    = bdry_variablenames,
                         static_variablenames  = static_variablenames)
  display(C)

  # test with no registered variables, should print a warning
  C = dg1d.Cache(mesh)
  display(C)

end


@testset "broadcast_volume!" begin

  # some definitions
  struct MyEquation end

  # need an @eval here, because of nested @testsets
  @eval @with_signature function compute_rhs(eq::MyEquation)
    @accepts u, v
    rhs_u = 2.0 * u
    rhs_v = 3.0 * v - 1.0
    @returns rhs_u, rhs_v
  end

  # setup cache
  mesh = Mesh(N=5,K=5)
  cache = Cache(mesh)
  dg1d.register_variables!(cache,
        dynamic_variablenames = (:u, :v),
        rhs_variablenames = (:rhs_u, :rhs_v))

  # prepare data
  @unpack u, v         = get_dynamic_variables(cache)
  @unpack rhs_u, rhs_v = get_rhs_variables(cache)

  # fill u,v with random numbers
  map!.(el -> randn(), [u, v], [u, v])

  # compute
  broadcast_volume!(compute_rhs, MyEquation(), cache)

  expected_rhs_u = @. 2.0 * u
  expected_rhs_v = @. 3.0 * v - 1.0

  @test all(isapprox.(rhs_u, expected_rhs_u))
  @test all(isapprox.(rhs_v, expected_rhs_v))

  #### edge cases

  # call broadcast_volume! with function that was not defined with @with_signature
  @test_throws ErrorException broadcast_volume!(args -> args, MyEquation(), cache)
  # call broadcast_volume! with function and wrong dispatch type
  struct WrongEquation end
  @test_throws ErrorException broadcast_volume!(args -> args, WrongEquation(), cache)

end


end
